"""
Tools for helping with :mod:`attrs`, particularly validators

TODO: move this into its own package and check with attrs developers
if they want something like this in their package
"""
from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, TypeVar

    import attr

    T = TypeVar("T")


class AttributeInitialisationError(ValueError):
    """
    Raised when there is an issue while initialising an :obj:`attr.Attribute`
    """

    def __init__(
        self,
        instance: Any,
        attribute: attr.Attribute[Any],
        value: T,
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        instance
            Instance being initialised

        attribute
            Attribute being set

        value
            Value being used to set the attribute
        """
        error_msg = (
            "Error raised while initialising attribute "
            f"``{attribute.name}`` of ``{type(instance)}``. "
            f"Value provided: {value}"
        )

        super().__init__(error_msg)


def add_attrs_context(
    original: Callable[[Any, attr.Attribute[Any], T], None],
) -> Callable[[Any, attr.Attribute[Any], T], None]:
    """
    Decorate function with a ``try...except`` to add the :mod:`attrs` context

    This means that the information about what attribute was being set and
    what value it was passed is also shown to the user

    Parameters
    ----------
    original
        Function to decorate

    Returns
    -------
        Decorated function

    Notes
    -----
    Only works with Python 3.11 and above. For other Python versions, the raw
    error is simply shown instead
    """

    @wraps(original)
    def with_attrs_context(
        instance: Any,
        attribute: attr.Attribute[Any],
        value: T,
    ) -> None:
        try:
            original(instance, attribute, value)
        except Exception as exc:
            raise AttributeInitialisationError(instance=instance, attribute=attribute, value=value) from exc

    return with_attrs_context


def make_attrs_validator_compatible_single_input(
    func_to_wrap: Callable[[T], None],
) -> Callable[[Any, attr.Attribute[Any], T], None]:
    """
    Create a function that is compatible with validation via :func:`attrs.field`

    This assumes that the function you're wrapping only takes a single input.

    Parameters
    ----------
    func_to_wrap
        Function to wrap

    Returns
    -------
        Wrapped function, which can be used as a validator with
        :func:`attrs.field`
    """

    @add_attrs_context
    @wraps(func_to_wrap)
    def attrs_compatible(
        instance: Any,
        attribute: attr.Attribute[Any],
        value: T,
    ) -> None:
        func_to_wrap(value)

    return attrs_compatible



"""
Generation of tasks for copying source into the outputs
"""
from __future__ import annotations

import json
import shutil
from collections.abc import Callable, Iterable
from functools import partial
from pathlib import Path
from typing import Any, Protocol

from attrs import frozen

from .doit_tools import swallow_output
from .typing import DoitTaskSpec


@frozen
class ActionDef:
    """Definition of an action"""

    name: str
    """Name of the action"""

    action: tuple[Callable[..., Any], list[Any], dict[str, Any]]
    """Action to execute with doit"""

    targets: tuple[Path, ...]
    """Files that this action creates"""


class CopyReadmeCallable(Protocol):
    """Callable which can be used for copying READMEs"""

    def __call__(self, in_path: Path, out_path: Path, run_id: str, config_file_raw: Path) -> None:
        """
        Copy README into the output bundle

        Parameters
        ----------
        in_path
            Path from which to copy the README

        out_path
            Path to copy the README to

        run_id
            ID of the run.

        config_file_raw
            Path to the raw configuration file
        """
        ...  # pragma: no cover


class CopyZenodoCallable(Protocol):
    """Callable which can be used for copying Zenodo JSON files"""

    def __call__(self, in_path: Path, out_path: Path, version: str) -> None:
        """
        Copy Zenodo JSON file into the output bundle

        Parameters
        ----------
        in_path
            Path from which to copy the README

        out_path
            Path to copy the README to

        version
            Version to write in the Zenodo file
        """
        ...  # pragma: no cover


class CopyFileCallable(Protocol):
    """Callable which can be used for copying files"""

    def __call__(self, in_path: Path, out_path: Path) -> None:
        """
        Copy a file from one location to another

        Parameters
        ----------
        in_path
            File to copy

        out_path
            Location to copy to
        """
        ...  # pragma: no cover


class CopyTreeCallable(Protocol):
    """Callable which can be used for copying file trees"""

    def __call__(self, in_path: Path, out_path: Path) -> None:
        """
        Copy a file from one location to another

        Parameters
        ----------
        in_path
            Tree to copy from

        out_path
            Location to copy to
        """
        ...  # pragma: no cover


def gen_copy_source_into_output_tasks(  # noqa: PLR0913
    all_preceeding_tasks: Iterable[DoitTaskSpec],
    repo_root_dir: Path,
    root_dir_output_run: Path,
    run_id: str,
    root_dir_raw_notebooks: Path,
    config_file_raw: Path,
    readme: str = "README.md",
    zenodo: str = "zenodo.json",
    other_files_to_copy: tuple[str | Path, ...] = (
        "dodo.py",
        "poetry.lock",
        "pyproject.toml",
    ),
    src_dir: str = "src",
    copy_readme: CopyReadmeCallable | None = None,
    copy_zenodo: CopyZenodoCallable | None = None,
    copy_file: CopyFileCallable | None = None,
    copy_tree: CopyTreeCallable | None = None,
) -> Iterable[DoitTaskSpec]:
    """
    Generate tasks to copy the source into the output directory

    Parameters
    ----------
    all_preceeding_tasks
        All tasks preceeding this one. The targets of these tasks are set
        as dependencies of this task to ensure that this task runs after them.

    repo_root_dir
        Root directory of the repository. This is used to know where to copy
        files from.

    root_dir_output_run
        Root directory of the run's output. This is used to know where to copy
        files to.

    run_id
        ID of the run.

    root_dir_raw_notebooks
        Root directory to the raw notebooks (these are copied into the output
        bundle to ensure that the bundle can be run standalone)

    config_file_raw
        Path to the raw configuration file

    config_file_raw_output_name
        Name to use when saving the raw configuration file to the output bundle
        (must be different to the hydrated configuration file to avoid a clash)

    readme
        Name of the README file to copy into the output

    zenodo
        Name of the zenodo JSON file to copy into the output

    other_files_to_copy
        Other files to copy into the output (paths are assumed to be relative
        to the project's root)

    src_dir
        Path to the Python source (this is also copied into the output bundle)

    copy_readme
        Callable that copies the README file into the output bundle. If not supplied,
        we use :func:`copy_readme_default`.

    copy_zenodo
        Callable that copies the Zenodo JSON file into the output bundle. If not supplied,
        we use :func:`copy_zenodo_default`.

    copy_file
        Callable that copies a file from one location to another. If not supplied, we use
        :func:`copy_file_default`.

    copy_tree
        Callable that copies a file tree from one location to another. If not supplied, we use
        :func:`copy_tree_default`.

    Yields
    ------
        Tasks for copying the source files into the output directory
    """
    if copy_readme is None:
        copy_readme = copy_readme_default

    if copy_zenodo is None:
        copy_zenodo = copy_zenodo_default

    if copy_file is None:
        copy_file = copy_file_default

    if copy_tree is None:
        copy_tree = copy_tree_default

    all_targets = []
    for task in all_preceeding_tasks:
        if "targets" in task:
            all_targets.extend(task["targets"])

    base_task = {
        "basename": "copy_source_into_output",
        "doc": (
            "Copy required source files into the output directory, making it "
            "easy to create a neat bundle for uploading to Zenodo"
        ),
    }

    output_dir_raw_notebooks = root_dir_output_run / root_dir_raw_notebooks.relative_to(repo_root_dir)

    config_file_raw_output = root_dir_output_run / f"{config_file_raw.stem}-raw{config_file_raw.suffix}"

    action_defs = [
        ActionDef(
            name="copy README",
            action=(
                copy_readme,
                [
                    repo_root_dir / readme,
                    root_dir_output_run / readme,
                    run_id,
                    config_file_raw_output.relative_to(root_dir_output_run),
                ],
                {},
            ),
            targets=(root_dir_output_run / readme,),
        ),
        ActionDef(
            name="copy Zenodo",
            action=(
                copy_zenodo,
                [repo_root_dir / zenodo, root_dir_output_run / zenodo, run_id],
                {},
            ),
            targets=(root_dir_output_run / zenodo,),
        ),
        *get_copy_file_action_definitions(
            repo_root_dir=repo_root_dir,
            root_dir_output_run=root_dir_output_run,
            other_files_to_copy=other_files_to_copy,
            copy_file=copy_file,
        ),
        ActionDef(
            name="copy raw config",
            action=(
                copy_file,
                [config_file_raw, config_file_raw_output],
                {},
            ),
            targets=(config_file_raw_output,),
        ),
        ActionDef(
            name="copy raw notebooks",
            action=(
                copy_tree,
                [root_dir_raw_notebooks, output_dir_raw_notebooks],
                {},
            ),
            targets=(output_dir_raw_notebooks,),
        ),
        ActionDef(
            name="copy source",
            action=(
                copy_tree,
                [repo_root_dir / src_dir, root_dir_output_run / src_dir],
                {},
            ),
            targets=(root_dir_output_run / src_dir,),
        ),
    ]

    for action_def in action_defs:
        created_files_short = tuple(f".../{t.name}" for t in action_def.targets)
        yield {
            "basename": base_task["basename"],
            "doc": f"{base_task['doc']}. Copying in {created_files_short}",
            "name": action_def.name,
            "actions": [action_def.action],
            "targets": action_def.targets,
            "file_dep": all_targets,
        }


def get_copy_file_action_definitions(
    repo_root_dir: Path,
    root_dir_output_run: Path,
    other_files_to_copy: tuple[str | Path, ...],
    copy_file: CopyFileCallable,
) -> Iterable[ActionDef]:
    """
    Get action definitions for copying other files

    Parameters
    ----------
    repo_root_dir
        Root directory of the repository. This is used to know where to copy
        files from.

    root_dir_output_run
        Root directory of the run's output. This is used to know where to copy
        files to.

    other_files_to_copy
        Other files to copy into the output (paths are assumed to be relative
        to the project's root)

    copy_file
        Callable that copies a file from one location to another.

    Returns
    -------
        Action definitions for copying the files

    Raises
    ------
    ValueError
        Any of the values in ``other_files_to_copy`` is an absolute path. Paths
        in ``other_files_to_copy`` are assumed to be relative to ``repo_root_dir``
        hence must be relative.
    """
    copy_file_action_definitions = []
    for file in other_files_to_copy:
        if isinstance(file, Path):
            if file.is_absolute():
                msg = (
                    f"{file} is absolute. "
                    "`other_files_to_copy` must not contain absolute paths "
                    "(all paths are assumed to be relative to `repo_root_dir`)."
                )
                raise ValueError(msg)

        copy_file_action_definitions.append(
            ActionDef(
                name=f"copy {file}",
                action=(
                    copy_file,
                    [repo_root_dir / file, root_dir_output_run / file],
                    {},
                ),
                targets=(root_dir_output_run / file,),
            )
        )

    return copy_file_action_definitions


class GetPydoitNBRunCommandCallable(Protocol):
    """Callable that supports inference of the run command to use with the pydoit-nb output bundle"""

    def __call__(self, config_file_raw: Path, raw_run_instruction: str) -> str:
        """
        Get run command

        This is the default implementation, based on how we have typically set up
        our pydoit-nb based repositories.

        Parameters
        ----------
        config_file_raw
            Path to raw configuration file as it should be used in the run command

        raw_run_instruction
            The command to run the analysis in the README (before the pydoit
            specific instructions are added).

        Returns
        -------
            Run command for use in the pydoit-nb created bundle
        """
        ...  # pragma: no cover


def copy_readme_default(  # noqa: PLR0913
    in_path: Path,
    out_path: Path,
    run_id: str,
    config_file_raw: Path,
    raw_run_instruction: str = "poetry run doit run --verbosity=2",
    get_pydoit_nb_run_command: GetPydoitNBRunCommandCallable | None = None,
) -> None:
    """
    Copy the README into the output bundle

    This is the default implementation. If you need other behaviour (for
    example, a different footer), write your own function then inject it as
    needed (e.g., into
    :func:`gen_copy_source_into_output_tasks`).

    Parameters
    ----------
    in_path
        Path to the raw README file (normally in the repository's root
        directory)

    out_path
        Path in which to write the README file (normally in the output bundle)

    run_id
        ID of the run. This is injected into the written README as part of the
        footer.

    config_file_raw
        Path to the raw configuration file, relative to the root output
        directory

    raw_run_instruction
        Instructions for how to run the workflow as they appear in the README.
        These are included to check that the instructions for running in the
        bundle are (likely) correct.

    get_pydoit_nb_run_command
        Callable that can create the run command to put in the pydoit-nb
        specific section of the copied README (this often differs from
        the command that is in the README to handle different paths,
        configuration files etc.).

    Raises
    ------
    AssertionError
        The run instructions in the README aren't as expected hence the code
        injection likely won't work as expected.

    ValueError
        ``config_file_raw`` is an absolute path. It should always be a relative
        path to ensure portability.
    """
    if get_pydoit_nb_run_command is None:
        get_pydoit_nb_run_command = get_pydoit_nb_run_command_default

    if config_file_raw.is_absolute():
        msg = f"``config_file_raw`` must be a relative path, received: {config_file_raw}"
        raise ValueError(msg)

    with open(in_path) as fh:
        raw = fh.read()

    if raw_run_instruction not in raw:
        msg = (
            "Could not find expected run instructions in README. "
            "The injected run instructions probably won't be correct. "
            f"Expected run instruction: {raw_run_instruction}"
        )
        raise AssertionError(msg)

    footer = f"""
## Pydoit info

This README was created from the raw {in_path.name} file as part of the {run_id!r} run with
[pydoit](https://pydoit.org/contents.html). The bundle should contain
everything required to reproduce the outputs. The environment can be
made with [poetry](https://python-poetry.org/) using the `poetry.lock` file
and the `pyproject.toml` file. Please disregard messages about the `Makefile`
in this file.

If you are looking to re-run the analysis, then you should run the below

```sh
{get_pydoit_nb_run_command(config_file_raw, raw_run_instruction)}
```

The reason for this is that you want to use the configuration with relative
paths. The configuration file that is included in this output bundle contains
absolute paths for reproducibility and traceability reasons. However, such
absolute paths are not portable so you need to use the configuration file
above, which is the raw configuration file as it was was used in the original
run.

If you have any issues running the analysis, please make an issue in our code
repository or reach out via email.
"""
    with open(out_path, "w") as fh:
        fh.write(raw)
        fh.write(footer)


def get_pydoit_nb_run_command_default(config_file_raw: Path, raw_run_instruction: str) -> str:
    """
    Get run command

    This is the default implementation, based on how we have typically set up
    our pydoit-nb based repositories.

    Parameters
    ----------
    config_file_raw
        Path to raw configuration file as it should be used in the run command

    raw_run_instruction
        The command to run the analysis in the README (before the pydoit
        specific instructions are added).

    Returns
    -------
        Run command for use in the pydoit-nb created bundle
    """
    return f"DOIT_CONFIGURATION_FILE={config_file_raw} {raw_run_instruction}"


def copy_zenodo_default(in_path: Path, out_path: Path, version: str) -> None:
    """
    Copy Zenodo JSON file to the output bundle

    This is the default implementation that updates the version information
    too. If you need other behaviour, write your own function then inject it as
    needed (e.g., into :func:`gen_copy_source_into_output_tasks`).

    More information about the metadata that can be uploaded via the zenodo
    API can be found under the "Deposit metadata" heading in the
    `zenodo API docs <https://developers.zenodo.org/#representation>`_.

    Parameters
    ----------
    in_path
        Path to raw Zenodo file

    out_path
        Path to output Zenodo file in the bundle

    version
        Version to write in the Zenodo file
    """
    with open(in_path) as fh:
        zenodo_metadata = json.load(fh)

    zenodo_metadata["metadata"]["version"] = version

    with open(out_path, "w") as fh:
        fh.write(json.dumps(zenodo_metadata, indent=2))


copy_file_default = swallow_output(shutil.copy2)

copy_tree_default = swallow_output(
    partial(
        shutil.copytree,
        ignore=shutil.ignore_patterns("*.pyc", "__pycache__"),
        dirs_exist_ok=True,
    )
)



"""
Checklist file generation
"""
from __future__ import annotations

from collections.abc import Callable, Iterable
from pathlib import Path

from doit.dependency import get_file_md5

CHECKLIST_FNAME_DEFAULT: str = "checklist.chk"
"""Default name used for checklist files"""


def get_checklist_file(directory: Path, filename: str = CHECKLIST_FNAME_DEFAULT) -> Path:
    """
    Get the full file path for a checklist

    Parameters
    ----------
    directory
        Directory for which we want to get the checklist file

    filename
        Name of the checklist file

    Returns
    -------
        Path of the checklist file

    """
    return directory / filename


def is_checklist_file(fp: Path, checklist_filename: str = CHECKLIST_FNAME_DEFAULT) -> bool:
    """
    Check if a file is a checklist file

    Parameters
    ----------
    fp
        The file to check

    checklist_file_name
        If the filename matches this value, it will be considered
        a checklist file.

    Returns
    -------
        ``True`` if the file is a checklist file, otherwise ``False``
    """
    return fp.name == checklist_filename


def create_md5_dict(
    files: Iterable[Path],
    exclusions: Iterable[Callable[[Path], bool]] | None = None,
) -> dict[Path, str]:
    """
    Create dictionary of MD5 hashes for files

    Parameters
    ----------
    files
        Files to create hashes for

    exclusions
        An iterable of callables. These are applied to each file. If any of the
        results is ``True`` then the file is skipped and will not be included
        in the dictionary of calculated hashes.

    Returns
    -------
        Dictionary of MD5 hashes for each file which was not excluded. The keys
        are the file paths (as they appear in ``files``) and the values are
        the calculated MD5 hashes.
    """
    out = {}
    for fp in files:
        if exclusions is not None and any(excl(fp) for excl in exclusions):
            # Don't include this file
            continue

        out[fp] = get_file_md5(fp)

    return out


def generate_directory_checklist(
    directory: Path,
    checklist_file: Path | None = None,
    exclusions: Iterable[Callable[[Path], bool]] = (is_checklist_file,),
) -> Path:
    """
    Create a checklist file with checksums for all files in a directory

    Running this command multiple times should result in the same result. This
    enables the checklist to be used as a target for doit tasks.

    The resulting checklist file can also be used to verify the contents
    of a folder using the program `mdfsum` so can be included in any
    distributed results.

    .. code:: bash

        md5sum -c checklist.chk

    Parameters
    ----------
    directory
        Directory containing arbitary files (we haven't tested this on any file
        but any directory containing hashable data files is the intended target)

    checklist_file
        Where to write the checklist file. If not supplied, the result of
        `get_checklist_file(directory)` is used.

    exclusions
        Functions used to check if a file should be excluded or not.

    Returns
    -------
        Path of the generated checklist file

    Raises
    ------
    NotADirectoryError
        If ``directory`` doesn't exist or isn't a directory
    """
    if not directory.is_dir():
        raise NotADirectoryError(directory)

    if checklist_file is None:
        checklist_file = get_checklist_file(directory)

    # sort to ensure same result for same set of files
    files = sorted([f for f in directory.rglob("*") if f.is_file()])

    md5s = create_md5_dict(files, exclusions=exclusions)

    with open(checklist_file, "w") as fh:
        for fp, md5 in md5s.items():
            fh.write(f"{md5}  {fp.relative_to(directory)}\n")

    return checklist_file



"""
Helper functions that can be useful when setting up config
"""
from __future__ import annotations

import collections
from collections.abc import Collection
from pathlib import Path
from typing import TYPE_CHECKING

from pydoit_nb.attrs_helpers import add_attrs_context
from pydoit_nb.config_handling import get_step_config_ids
from pydoit_nb.typing import NotebookConfigLike

if TYPE_CHECKING:
    from typing import Any

    import attr


def assert_path_exists(inp: Path) -> None:
    """
    Assert that a path exists

    Parameters
    ----------
    inp
        Path to check

    Raises
    ------
    FileNotFoundError
        ``inp`` does not exist
    """
    if not inp.exists():
        raise FileNotFoundError(f"{inp} does not exist")  # noqa: TRY003


def assert_path_is_absolute(inp: Path) -> None:
    """
    Assert that a path is an absolute path

    Parameters
    ----------
    inp
        Path to check

    Raises
    ------
    AssertionError
        ``inp`` is not absolute
    """
    if not inp.is_absolute():
        raise AssertionError(f"{inp} is not absolute")  # noqa: TRY003


def assert_step_config_ids_are_unique(inp: Collection[NotebookConfigLike]) -> None:
    """
    Assert that all the :attr:`step_config_id` in a collection are unique

    Parameters
    ----------
    inp
        Collection of :obj:`NotebookConfigLike` objects to check.

    Raises
    ------
    AssertionError
        The :attr:`step_config_id` of each :obj:`NotebookConfigLike` are not unique.
    """
    step_config_ids = get_step_config_ids(inp)

    if len(set(step_config_ids)) != len(inp):
        duplicates = [id for id, count in collections.Counter(step_config_ids).items() if count > 1]

        msg = (
            "``step_config_id`` must be unique. "
            f"The following ``step_config_id`` are duplicated: {duplicates}"
        )
        raise AssertionError(msg)


@add_attrs_context
def assert_path_is_subdirectory_of_root_dir_output(
    instance: Any, attribute: attr.Attribute[Any], value: Path
) -> None:
    """
    Assert that a path is a sub-directory of ``instance.root_dir_output``

    Parameters
    ----------
    instance
        Instance to check

    attribute
        Attribute that is being set

    value
        Value that is being used to set the attribute

    Raises
    ------
    AssertionError
        ``value`` is not a sub-directory of ``instance.root_dir_output``
    """
    if not value.is_relative_to(instance.root_dir_output):
        msg = (
            f"``{attribute.name}`` is not a sub-directory of root_dir_output. "
            f"{attribute.name}={value!r}. root_dir_output={instance.root_dir_output!r}"
        )
        raise AssertionError(msg)



"""
Notebook defining classes
"""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

from attrs import frozen
from doit.tools import config_changed  # type: ignore

from .notebook_run import run_notebook

if TYPE_CHECKING:
    from .typing import Converter, DoitTaskSpec

C = TypeVar("C")


@frozen
class UnconfiguredNotebook:
    """A notebook without any configuration"""

    notebook_path: Path
    """Path to notebook, relative to the raw notebook directory"""

    raw_notebook_ext: str
    """Extension for the raw notebook"""

    summary: str
    """One line summary of the notebook"""
    # TODO: validation?

    doc: str
    """Documentation of the notebook (can be longer than one line)"""


@frozen
class ConfiguredNotebook:
    """
    A configured notebook

    It might make sense to refactor this so has an UnconfiguredNotebook
    as one of its attributes rather than duplicatinng things. That would
    probably also make it clearer that a configured notebook is just the
    combination of an unconfigured notebook and the config we want to use
    with it.
    """

    unconfigured_notebook: UnconfiguredNotebook
    """
    The unconfigured notebook which should be combined with the configuration
    """

    dependencies: tuple[Path, ...]
    """Paths on which the notebook depends"""

    targets: tuple[Path, ...]
    """Paths which the notebook creates/controls"""

    config_file: Path
    """Path to the config file to use with the notebook"""

    step_config_id: str
    """`step_config_id` to use for this run of the notebook"""

    # Type hinting could be improved, but even attrs seems to just use Any
    # for the signature of `dumps` so I don't think it is worth diving down
    # that rabbit hole right now.
    configuration: tuple[Any, ...] | None = None
    """
    Configuration used by the notebook.

    If any of the configuration changes then the notebook will be triggered.

    If nothing is provided, then the notebook will be run whenever the
    configuration file driving the notebook is modified (i.e. the notebook will
    be re-run for any configuration change).
    """
    # TODO: It looks like this solves a problem that even the original authors
    # hadn't thought about because they just suggest using forget here
    # https://pydoit.org/cmd-other.html#forget (although they also talk about
    # non-file dependencies elsewhere so maybe these are just out of date docs)

    def to_doit_task(  # noqa: PLR0913
        self,
        root_dir_raw_notebooks: Path,
        notebook_output_dir: Path,
        base_task: DoitTaskSpec,
        converter: Converter | None = None,
        clean: bool = True,
        config_changed_class: Any = config_changed,
    ) -> DoitTaskSpec:
        """
        Convert to a :mod:`doit` task

        Parameters
        ----------
        root_dir_raw_notebooks
            Root directory in which the raw (not yet run) notebooks are kept

        notebook_output_dir
            Directory in which to write out the run notebook

        base_task
            Base task definition for this notebook step

        converter
            Converter to use to serialise configuration if needed.

        clean
            If we run `doit clean`, should the targets also be removed?

        config_changed_class
            Class to use for creating our object which can identify if config
            has changed or not. This should only need to be changed in very
            rare circumstances. The type hint is very vague because we don't
            want to repeat the pydoit API/logic here.

        Returns
        -------
            Task specification for use with :mod:`doit`

        Raises
        ------
        TypeError
            ``self.configuration is not None`` but ``converter is None``
        """
        raw_notebook = root_dir_raw_notebooks / self.unconfigured_notebook.notebook_path.with_suffix(
            self.unconfigured_notebook.raw_notebook_ext
        )

        notebook_name = self.unconfigured_notebook.notebook_path.name
        unexecuted_notebook = notebook_output_dir / f"{notebook_name}_unexecuted.ipynb"
        executed_notebook = notebook_output_dir / f"{notebook_name}.ipynb"

        dependencies = [
            *self.dependencies,
            raw_notebook,
        ]
        notebook_parameters = dict(config_file=str(self.config_file), step_config_id=self.step_config_id)

        targets = self.targets

        task = dict(
            basename=base_task["basename"],
            name=self.step_config_id,
            doc=f"{base_task['doc']}. step_config_id={self.step_config_id!r}",
            actions=[
                (
                    run_notebook,
                    [],
                    {
                        "raw_notebook": raw_notebook,
                        "unexecuted_notebook": unexecuted_notebook,
                        "executed_notebook": executed_notebook,
                        "notebook_parameters": notebook_parameters,
                    },
                )
            ],
            targets=targets,
            file_dep=dependencies,
            clean=clean,
        )

        if self.configuration is None:
            # Run whenever config file changes
            task["file_dep"].extend([self.config_file])
        else:
            if converter is None:
                msg = "If `self.configuration is not None` then `converter` must be supplied"
                raise ValueError(msg)

            has_config_changed = config_changed_class(converter.dumps(self.configuration, sort_keys=True))

            task["uptodate"] = (has_config_changed,)

        return task



"""
Notebook-based step

A notebook-based step is the combination of a notebook and the configuration
to run it.
"""
from __future__ import annotations

import copy
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar

from attrs import frozen

from pydoit_nb.config_handling import get_step_config_ids

from .typing import ConfigBundleLike, Converter, DoitTaskSpec

if TYPE_CHECKING:
    from .notebook import ConfiguredNotebook, UnconfiguredNotebook


C = TypeVar("C")
CB_contra = TypeVar("CB_contra", contravariant=True, bound=ConfigBundleLike[Any])


class ConfigureNotebooksCallable(Protocol[CB_contra]):
    """Callable that can be used for configuring notebooks"""

    def __call__(  # noqa: D102
        self,
        unconfigured_notebooks: Iterable[UnconfiguredNotebook],
        config_bundle: CB_contra,
        step_name: str,
        step_config_id: str,
    ) -> list[ConfiguredNotebook]:
        ...  # pragma: no cover


@frozen
class UnconfiguredNotebookBasedStep(Generic[C, CB_contra]):
    """
    An unconfigured notebook-based step

    A step is a step in the overall workflow. A notebook-based step can be made
    up of one or more notebooks. These are then configured at run-time with the
    run-time information so they can then be turned into doit task(s).
    """

    step_name: str
    """Name of the step"""

    unconfigured_notebooks: list[UnconfiguredNotebook]
    """Unconfigured notebooks that make up this step"""

    configure_notebooks: ConfigureNotebooksCallable[CB_contra]
    """Function which can configure the notebooks based on run-time information"""

    def gen_notebook_tasks(
        self,
        config_bundle: CB_contra,
        root_dir_raw_notebooks: Path,
        converter: Converter | None = None,
        clean: bool = True,
    ) -> Iterable[DoitTaskSpec]:
        """
        Generate notebook tasks for this step

        Parameters
        ----------
        config_bundle
            Configuration bundle to use when generating the tasks

        root_dir_raw_notebooks
            Root directory in which the raw notebooks live

        converter
            Instance that can serialise the configuration used by each notebook

        clean
            If we run `doit clean`, should the targets of each task be
            removed?

        Yields
        ------
            Task specifications for use with :mod:`doit`
        """
        unconfigured_notebooks = self.unconfigured_notebooks

        unconfigured_notebooks_base_tasks = {}
        for nb in unconfigured_notebooks:
            base_task = {
                "basename": f"({nb.notebook_path}) {nb.summary}",
                "name": None,
                "doc": nb.doc,
            }
            # yield copy of base task to avoid being mangled by doit
            yield copy.deepcopy(base_task)

            unconfigured_notebooks_base_tasks[nb.notebook_path] = base_task

        step_config_ids = get_step_config_ids(getattr(config_bundle.config_hydrated, self.step_name))

        notebook_output_dir_step = config_bundle.root_dir_output_run / "notebooks-executed" / self.step_name
        for step_config_id in step_config_ids:
            configured_notebooks = self.configure_notebooks(
                unconfigured_notebooks,
                config_bundle=config_bundle,
                step_name=self.step_name,
                step_config_id=step_config_id,
            )
            if len(unconfigured_notebooks) != len(configured_notebooks):
                msg = (
                    "The number of unconfigured and configured notebooks is not the same. "
                    "We haven't yet thought through this use case. "
                    "Please raise an issue at https://github.com/climate-resource/pydoit-nb to discuss."
                )
                raise NotImplementedError(msg)

            notebook_output_dir_step_id = notebook_output_dir_step / step_config_id

            for nb_configured in configured_notebooks:
                notebook_task = nb_configured.to_doit_task(
                    root_dir_raw_notebooks=root_dir_raw_notebooks,
                    notebook_output_dir=notebook_output_dir_step_id,
                    base_task=unconfigured_notebooks_base_tasks[
                        nb_configured.unconfigured_notebook.notebook_path
                    ],
                    converter=converter,
                    clean=clean,
                )

                yield notebook_task



"""
Re-usable tools for generating tasks
"""
from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from typing import Any, Callable, Protocol, TypeVar

from pydoit_nb.notebook_step import UnconfiguredNotebookBasedStep
from pydoit_nb.typing import ConfigBundleLike, Converter, DoitTaskSpec

C = TypeVar("C")
CB_contra = TypeVar("CB_contra", bound=ConfigBundleLike[Any], contravariant=True)


class StepDefiningModuleLike(Protocol[C, CB_contra]):
    """
    Class that behaves like a module which defines a notebook-based step
    """

    @property
    def step(self) -> UnconfiguredNotebookBasedStep[C, CB_contra]:
        """
        Step this module defines
        """
        ...  # pragma: no cover


def generate_all_tasks(
    config_bundle: ConfigBundleLike[C],
    root_dir_raw_notebooks: Path,
    converter: Converter,
    step_defining_modules: Iterable[StepDefiningModuleLike[C, ConfigBundleLike[C]]],
    gen_zenodo_bundle_task: Callable[[list[DoitTaskSpec]], DoitTaskSpec],
) -> Iterable[DoitTaskSpec]:
    """
    Generate all tasks in the workflow

    This is a helper function for a common pattern. You can obviously do
    whatever you want in your own projects as there is a lot of flexibility.

    Parameters
    ----------
    config_bundle
        Configuration bundle to use in task generation

    root_dir_raw_notebooks
        Directory in which raw notebooks are kept.

    converter
        Object that can serialize the configuration bundle's hydrated config

    step_defining_modules
        Modules that define notebook-based steps

    gen_zenodo_bundle_task
        Callable that takes all the proceeding tasks and returns a task for
        generating a bundle which can be uploaded to Zenodo.

    Yields
    ------
        :mod:`doit` tasks to run
    """
    notebook_tasks: list[DoitTaskSpec] = []
    for step_defining_module in step_defining_modules:
        for task in step_defining_module.step.gen_notebook_tasks(
            config_bundle=config_bundle,
            root_dir_raw_notebooks=root_dir_raw_notebooks,
            converter=converter,
        ):
            yield task
            notebook_tasks.append(task)

    yield gen_zenodo_bundle_task(notebook_tasks)



"""
Tools for compatibility with [doit](pydoit.org)
"""
from __future__ import annotations

import functools
import logging
from collections.abc import Callable
from typing import Any


def setup_logging(
    stdout_level: int = logging.WARNING,
    log_file: str | None = "dodo.log",
    file_level: int = logging.DEBUG,
    log_fmt: str = (
        "%(levelname)s - %(asctime)s %(name)s %(processName)s "
        "(%(module)s:%(funcName)s:%(lineno)d):  %(message)s"
    ),
    datefmt: str = "%Y-%m-%d %H:%M:%S",
) -> logging.Logger:
    """
    Set up logging

    This is a conveniance function. It does things like adding handlers to the
    root logger, which won't always be helpful or desired so use with some care.
    It will not cover all possible logging use cases. If you have a more
    complicated use case, you will probably need to write your own set up
    implementation. However, this implementation may be of some use.

    Parameters
    ----------
    stdout_level
        Level to use for logging to stdout

    log_file
        File to write on disk logs to

    file_level
        Level to use for logging to file

    log_fmt
        Format to use for log strings

    datefmt
        Format to use for dates

    Returns
    -------
        dodo logger
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    logFormatter = logging.Formatter(
        log_fmt,
        datefmt=datefmt,
    )

    stdoutHandler = logging.StreamHandler()
    stdoutHandler.setFormatter(logFormatter)
    stdoutHandler.setLevel(stdout_level)

    root_logger.addHandler(stdoutHandler)

    if log_file is not None:
        fileHandler = logging.FileHandler(log_file)
        fileHandler.setFormatter(logFormatter)
        fileHandler.setLevel(file_level)

        root_logger.addHandler(fileHandler)

    return logging.getLogger("dodo")


def swallow_output(func: Callable[..., Any]) -> Callable[..., None]:
    """
    Decorate function so the output is swallowed

    This is needed to make pydoit recognise the task has run correctly

    Parameters
    ----------
    func
        Function to decorate

    Returns
    -------
        Decorated function
    """

    @functools.wraps(func)
    def out(*args: Any, **kwargs: Any) -> None:
        func(*args, **kwargs)

    return out



"""
Typing specifications
"""
from __future__ import annotations

from pathlib import Path
from typing import Any, Protocol, TypeVar

from attrs import AttrsInstance

try:
    from typing_extensions import TypeAlias
except ImportError:  # >= python 3.11
    # remove type ignore when mypy applied with python 3.11
    from typing import TypeAlias  # type: ignore

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)

DoitTaskSpec: TypeAlias = dict[str, Any]


class ConfigBundleLike(Protocol[T_co]):
    """
    Protocol for configuration bundles
    """

    @property
    def config_hydrated(self) -> T_co:
        """Hydrated config"""
        ...

    @property
    def config_hydrated_path(self) -> Path:
        """Path in which to write the hydrated config"""
        ...

    @property
    def root_dir_output_run(self) -> Path:
        """Root directory in which output is saved"""
        ...

    @property
    def run_id(self) -> str:
        """Run ID for the run to which this config bundle applies"""
        ...


AI_contra = TypeVar("AI_contra", bound=AttrsInstance, contravariant=True)
CB_co = TypeVar("CB_co", bound=ConfigBundleLike[Any], covariant=True)


class ConfigBundleCreator(Protocol[AI_contra, CB_co]):
    """
    Callable that can be used to create config bundles
    """

    def __call__(
        self, config_hydrated: AI_contra, config_hydrated_path: Path, root_dir_output_run: Path
    ) -> CB_co:
        """
        Create :obj:`ConfigBundleLike`

        Parameters
        ----------
        config_hydrated
            Hydrated config to include in the bundle

        config_hydrated_path
            Path where the hydrated config is saved to disk

        root_dir_output_run
            Root directory for the run's output

        Returns
        -------
            Created :obj:`ConfigBundleLike`
        """
        ...  # pragma: no cover


class Converter(Protocol):
    """
    Protocol for converters
    """

    def dumps(self, obj: Any, sort_keys: bool = False) -> str:
        """
        Dump configuration to a string

        Parameters
        ----------
        obj
            Object to dump. The type hints aren't great here. The assumption is
            that the dumping protocol should handle any type issues (I think
            static typing doesn't really work here, for reasons I don't fully
            have my head around).

        sort_keys
            Should the keys be sorted in the output?

        Returns
        -------
            String version of ``config``
        """
        ...  # pragma: no cover

    def loads(self, inp: str, target: type[T]) -> T:
        """
        Load an instance of ``target`` from a string

        Parameters
        ----------
        inp
            String to load from

        target
            Object type to return

        Returns
        -------
            Loaded instance of ``target``
        """
        ...  # pragma: no cover


class NotebookConfigLike(Protocol):
    """
    A class which is like a notebook config
    """

    step_config_id: str
    """String which identifies the step config to use with the notebook"""



"""
Utilities for displaying things, typically by printing
"""
from __future__ import annotations

from pathlib import Path
from typing import Any

from pydoit_nb.typing import DoitTaskSpec


def print_config(**kwargs: Any) -> None:
    """
    Print configuration

    Parameters
    ----------
    **kwargs
        Config to show
    """
    config_str = "\n".join([f"\t{k}: {v}" for k, v in kwargs.items()])
    print(f"Will run with the following config:\n{config_str}\n")


def gen_show_configuration_task(
    configuration_file: Path,
    run_id: str,
    root_dir_output: Path,
    root_dir_raw_notebooks: Path,
) -> DoitTaskSpec:
    """
    Generate a :mod:`doit` task that shows the configuration being used for the run

    Parameters
    ----------
    configuration_file
        Configuration file being used for the run

    run_id
        Run ID being used for the run

    root_dir_output
        Root directory in which the run's output will be saved

    root_dir_raw_notebooks
        Root directory from which the raw notebooks will be fetched

    Returns
    -------
        :mod:`doit` task that shows the configuration
    """
    return {
        "name": "Show configuration",
        "actions": [
            (
                print_config,
                [],
                dict(
                    configuration_file=configuration_file,
                    run_id=run_id,
                    root_dir_output=root_dir_output,
                    root_dir_raw_notebooks=root_dir_raw_notebooks,
                ),
            )
        ],
    }



"""
Complete file generation
"""
from __future__ import annotations

import datetime as dt
from pathlib import Path


def write_complete_file(complete_file: Path, contents: str | None = None) -> Path:
    """
    Write a complete file

    Parameters
    ----------
    complete_file
        Path in which to write the complete file

    contents
        Contents to write in the file. If ``None``, we simply write a timestamp
        into the file.

    Returns
    -------
        Path to the written file
    """
    if contents is None:
        contents = dt.datetime.now().strftime("%Y%m%d%H%M%S")

    with open(complete_file, "w") as fh:
        fh.write(contents)

    return complete_file



"""
Serialization tools

This module includes a default yaml converter. This may be exactly what you
need for your use case. If it isn't, the code will help you see how to set up
and customise a converter. This example, along with the cattrs docs, should
help you get going. We mainly use :func:`cattrs.register_structure_hook_func`
and :func:`cattrs.register_unstructure_hook_func` (see also
`here <https://catt.rs/en/stable/cattrs.html#cattrs.BaseConverter.register_structure_hook_func>`_
), which aren't as heavily documented, but are what we want. You may find
the other examples in
`the docs on customisation <https://catt.rs/en/stable/customizing.html#customizing-class-un-structuring>`_
more helpful.

One example of something we don't support well is preservation of data types
when structuring and unstructuring pint types. This is just a complicated
problem to do with type hinting. For most applications, we have found that the
default numpy data type is fine (everything becomes float64, which is generally
fine).

One other clear example of something we don't support here is structuring and
unstructuring pandas objects as yaml. The reason is that we can't see an easy
way to get the round-tripping to work in edge cases related to index naming and
data type preservation. There are ways to solve this problem, but they require
more care and time than we can put in right now. As two suggestions for
possible solutions: 1) use a back-end that isn't yaml 2) store information
about index types etc. in addition to the data frame (pandas itself doesn't do
this though, which makes us a bit nervous about how hard it could be to do in
the general case, although specific use cases should be far more tractable and
easy to test). If you'd like to discuss this more, please raise an issue.
"""
from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Any, TypeVar, Union, cast

import cattrs.preconf.pyyaml
import numpy as np
import numpy.typing as nptype

from .typing import ConfigBundleLike, Converter

try:
    from typing_extensions import TypeAlias
except ImportError:  # >= python 3.11
    # remove type ignore when mypy applied with python 3.11
    from typing import TypeAlias  # type: ignore

try:
    import pint

    HAS_PINT = True
except ImportError:  # pragma: no cover
    HAS_PINT = False

T = TypeVar("T")
U = TypeVar("U")
N = TypeVar("N", bound=nptype.NDArray[Union[np.floating[Any], np.integer[Any]]])


def write_config_in_config_bundle_to_disk(
    config_bundle: ConfigBundleLike[U],
    converter: Converter,
) -> Path:
    """
    Write the configuration in a configuration bundle to disk

    The configuration is written in the path specified by
    ``config_bundle.config_hydrated_path``

    Parameters
    ----------
    config_bundle
        Configuration bundle to write to disk

    converter
        Object that can serialize the configuration bundle's hydrated config

    Returns
    -------
        Path in which ``config_bundle.config_hydrated`` was written
    """
    write_path = config_bundle.config_hydrated_path
    with open(write_path, "w") as fh:
        fh.write(converter.dumps(config_bundle.config_hydrated))

    return write_path


def load_config_from_file(config_file: Path, target: type[T], converter: Converter) -> T:
    """
    Load configuration from file

    Parameters
    ----------
    config_file
        File from which to load configuration

    target
        Class to load

    converter
        Converter to use to convert from ``config_file``'s contents to an
        instance of ``target``

    Returns
    -------
        Loaded instance of ``target``
    """
    with open(config_file) as fh:
        config = converter.loads(fh.read(), target)

    return config


converter_yaml = cattrs.preconf.pyyaml.make_converter()


UnstructuredArray: TypeAlias = Union[Sequence[Union[int, float]], Sequence["UnstructuredArray"]]


def unstructure_np_array(arr: N) -> UnstructuredArray:
    """
    Unstructure :obj:`npt.ArrayLike`

    This simply converts it to a list so is probably not very fast. However,
    this is just an example so could easily be optimised for production use if
    needed.

    Parameters
    ----------
    arr
        Array to unstructure

    Returns
    -------
        Unstructured array
    """
    return cast(UnstructuredArray, arr.tolist())


def structure_np_array(inp: UnstructuredArray, target_type: type[N]) -> N:
    """
    Structure :obj:`npt.ArrayLke`

    The inverse of :func:`unstructure_np_array`

    Parameters
    ----------
    inp
        Data to structure

    target_type
        Type the data should be returned as

    Returns
    -------
        Structured array
    """
    # Can't get mypy to behave, hence type ignore comments throughout here
    # TODO: push docs PR up into cattrs
    # See https://github.com/python-attrs/cattrs/issues/194#issuecomment-987341893
    target_dtype = target_type.__args__[1].__args__[0]  # type: ignore

    return np.array([target_dtype(row) for row in inp])  # type: ignore


def _is_np_array(inp: Any) -> bool:
    return inp is np.ndarray or (getattr(inp, "__origin__", None) is np.ndarray)


converter_yaml.register_unstructure_hook_func(_is_np_array, unstructure_np_array)
converter_yaml.register_structure_hook_func(_is_np_array, structure_np_array)


def unstructure_np_scalar(number: np.number[Any]) -> float | int:
    """
    Unstructure :obj:`np.number`

    This simply converts to a primative type.

    Parameters
    ----------
    number
        Number to unstructure

    Returns
    -------
        Unstructured number
    """
    if isinstance(number, np.floating):
        return float(number)

    return int(number)


def structure_np_scalar(inp: float | int, target_type: type[T]) -> T:
    """
    Structure :obj:`np.number`

    The inverse of :func:`unstructure_np_array`

    Parameters
    ----------
    inp
        Data to structure

    target_type
        Type the data should be returned as

    Returns
    -------
        Structured number
    """
    # Can't get mypy to behave here either
    return target_type(inp)  # type: ignore


def _is_np_scalar(inp: Any) -> bool:
    return issubclass(inp, np.number)


converter_yaml.register_unstructure_hook_func(_is_np_scalar, unstructure_np_scalar)
converter_yaml.register_structure_hook_func(_is_np_scalar, structure_np_scalar)

if HAS_PINT:
    UnstructuredPint: TypeAlias = Union[tuple[Union[int, float], str], tuple[UnstructuredArray, str]]

    def unstructure_pint(inp: pint.UnitRegistry.Quantity) -> UnstructuredPint:
        """
        Unstructure a :mod:`pint` quantity.

        Parameters
        ----------
        inp
            :obj:`pint.UnitRegistry.Quantity` to unstructure

        Returns
        -------
            Unstructured :obj:`pint.UnitRegistry.Quantity`
        """
        if _is_np_scalar(type(inp.magnitude)):
            return (unstructure_np_scalar(inp.magnitude), str(inp.units))

        if isinstance(inp.magnitude, float):
            return (inp.magnitude, str(inp.units))

        return (unstructure_np_array(inp.magnitude), str(inp.units))

    def structure_pint(
        inp: UnstructuredPint, target_type: type[pint.UnitRegistry.Quantity]
    ) -> pint.UnitRegistry.Quantity:
        """
        Structure :obj:`pint.UnitRegistry.Quantity`

        Parameters
        ----------
        inp
            Unstructured data

        target_type
            Type to create

        Returns
        -------
            Structured :obj:`pint.UnitRegistry.Quantity`
        """
        # pint not playing nice with mypy
        ur = pint.get_application_registry()  # type: ignore

        # Can't do dtype control until pint allows it again with e.g.
        # pint.Quantity[np.array[np.float64]]
        return ur.Quantity(np.array(inp[0]), inp[1])  # type: ignore

    def _is_pint(inp: Any) -> bool:
        # I don't love this way of checking, but I couldn't work out how to else to make it work
        return hasattr(inp, "units") & hasattr(inp, "magnitude") & ("pint" in str(inp))

    converter_yaml.register_unstructure_hook_func(_is_pint, unstructure_pint)
    converter_yaml.register_structure_hook_func(_is_pint, structure_pint)

else:  # pragma: no cover
    # TODO: decide whether lack of pint should raise a warning or not
    pass



"""
Tools to help with testing
"""
from __future__ import annotations

import copy
from typing import Any


def assert_doit_tasks_equal(res: dict[str, Any], exp: dict[str, Any]) -> None:
    """
    Assert that doit tasks are equal

    This works around the fact that :class:`doit.tools.config_changed` objects
    don't compare equal in an easy to use way.

    Parameters
    ----------
    res
        Task to check

    exp
        Expected value of the task
    """
    # Copy the tasks before proceeding so that we can pop without affecting the
    # original result
    res_copy = copy.deepcopy(res)
    exp_copy = copy.deepcopy(exp)

    # doit doesn't implement a helpful comparison method so we have to do some
    # of our own comparison magic here
    res_uptodate = res_copy.pop("uptodate")
    assert len(res_uptodate) == 1

    exp_uptodate = exp_copy.pop("uptodate")
    assert len(exp_uptodate) == 1

    for attr in ("config", "config_digest", "encoder"):
        assert getattr(res_uptodate[0], attr) == getattr(exp_uptodate[0], attr)

    assert res_copy == exp_copy



"""
Re-usable tools that can help when creating configuration

These are likely most helpful for workflows that make use of :mod:`pydoit-nb`.
"""
from __future__ import annotations

from attrs import frozen


@frozen
class URLSource:
    """
    Source information for downloading a source from a URL
    """

    url: str
    """URL to download from"""

    known_hash: str
    """Known hash for the downloaded file"""



"""
Notebook running
"""
from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import Any, Callable

import jupytext
import papermill

logger = logging.getLogger(__name__)


_PYTHON_ADD_NOTE_INTRODUCED: int = 11
"""Python3 minor version in which add_note was introduced"""


class NotebookExecutionException(Exception):
    """
    Raised when a notebook fails to execute for any reason
    """

    def __init__(self, exc: Exception, filename: Path):
        note = f"{filename} failed to execute. Original exception: {exc}"
        if sys.version_info[1] >= _PYTHON_ADD_NOTE_INTRODUCED:
            self.add_note(note)  # type: ignore # can be removed once we apply mypy with Python 3.11

        super().__init__(exc)


def rewrite_notebook_default(
    raw_notebook: Path,
    unexecuted_notebook: Path,
    fmt: str = "ipynb",
) -> None:
    """
    Re-write notebooks from raw to unexecuted

    This is our default implementation to use in :func:`run_notebook`.

    Parameters
    ----------
    raw_notebook
        Path from which to read the raw notebook

    unexecuted_notebook
        Path in which to write the unexecuted notebook

    fmt
        Format in which to write the unexecuted notebook
    """
    logger.info("Reading raw notebook with jupytext: %s", raw_notebook)
    notebook_jupytext = jupytext.read(raw_notebook)

    logger.info("Writing unexecuted notebook: %s", unexecuted_notebook)
    # TODO: consider whether this should be elsewhere
    unexecuted_notebook.parent.mkdir(parents=True, exist_ok=True)
    jupytext.write(
        notebook_jupytext,
        unexecuted_notebook,
        fmt=fmt,
    )


def run_notebook(  # noqa: PLR0913
    raw_notebook: Path,
    unexecuted_notebook: Path,
    executed_notebook: Path,
    notebook_parameters: dict[str, str] | None = None,
    notebook_rewriter: Callable[[Path, Path], None] | None = None,
    notebook_executor: Callable[[Path, Path, dict[str, Any]], Any] | None = None,
) -> None:
    """
    Run a notebook

    This loads the notebook ``raw_notebook`` using jupytext, then writes it
    as an ``.ipynb`` file to ``unexecuted_notebook``. It then runs this
    unexecuted notebook with papermill, writing it to ``executed_notebook``.

    Parameters
    ----------
    raw_notebook
        Notebook from which to start

    unexecuted_notebook
        Where to write the unexecuted notebook

    executed_notebook
        Where to write the executed notebook

    notebook_parameters
        Parameters to pass to the target notebook

        These parameters will replace the contents of a cell tagged "parameters".
        See the
        `papermill documentation <https://papermill.readthedocs.io/en/latest/usage-parameterize.html#designate-parameters-for-a-cell>`_
        for more information about parameterizing a notebook.

    notebook_rewriter
        Function to use to re-write the raw notebooks as unexecuted notebooks
        that can be executed by `notebook_executor`. We use jupytext for this
        (specifically :func:`rewrite_notebook_default`) by default and you
        shouldn't need to change this.

    notebook_executor
        Function to use to execute the notebooks. We use `papermill.execute_notebook`
        for this by default and you shouldn't need to change this.
    """
    if notebook_parameters is None:
        notebook_parameters = {}

    if notebook_rewriter is None:
        notebook_rewriter = rewrite_notebook_default

    if notebook_executor is None:
        notebook_executor = papermill.execute_notebook

    notebook_rewriter(raw_notebook, unexecuted_notebook)

    try:
        if not executed_notebook.parent.exists():
            logger.info(
                "Creating directory (and any required parent directories): %s",
                executed_notebook.parent,
            )
            executed_notebook.parent.mkdir(parents=True, exist_ok=True)

        logger.info("Executing notebook: %s", unexecuted_notebook)
        notebook_executor(
            unexecuted_notebook,
            executed_notebook,
            notebook_parameters,
        )

    except Exception as exc:
        raise NotebookExecutionException(exc, unexecuted_notebook) from exc



"""
Tools for working with configuration
"""
from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from typing import Any, Callable, TypeVar, cast, overload

import attrs
import numpy as np
from attrs import AttrsInstance, evolve, fields

from pydoit_nb.serialization import write_config_in_config_bundle_to_disk
from pydoit_nb.typing import ConfigBundleCreator, ConfigBundleLike, Converter, NotebookConfigLike

T = TypeVar("T")


def insert_path_prefix(config: AI, prefix: Path) -> AI:
    """
    Insert path prefix into config attributes

    This adds the prefix ``prefix`` to any attributes of ``config`` which are
    :obj:`Path`

    Parameters
    ----------
    config
        Config to update

    prefix
        Prefix to add to paths

    Returns
    -------
        Updated ``config``
    """
    config_attrs = fields(config.__class__)

    evolutions: dict[str, Any] = {}
    for attr in config_attrs:
        attr_name = attr.name
        attr_value = getattr(config, attr_name)

        if isinstance(attr_value, dict):
            evolutions[attr_name] = {
                update_attr_value(k, prefix): update_attr_value(v, prefix) for k, v in attr_value.items()
            }

        elif not isinstance(attr_value, (str, np.ndarray)) and isinstance(attr_value, Iterable):
            evolutions[attr_name] = [update_attr_value(v, prefix) for v in attr_value]

        else:
            evolutions[attr_name] = update_attr_value(attr_value, prefix)

    return evolve(config, **evolutions)  # type: ignore # no idea why this fails


@overload
def update_attr_value(value: AttrsInstance, prefix: Path) -> AttrsInstance:
    ...  # pragma: no cover


@overload
def update_attr_value(value: Path, prefix: Path) -> Path:
    ...  # pragma: no cover


@overload
def update_attr_value(value: T, prefix: Path) -> T:
    ...  # pragma: no cover


def update_attr_value(value: AttrsInstance | Path | T, prefix: Path) -> AttrsInstance | Path | T:
    """
    Update the attribute value if it is :obj:`Path` to include the prefix

    The prefix is taken from the outer scope

    Parameters
    ----------
    value
        Value to update

    prefix
        Prefix to insert before paths if ``value`` is an instance of ``Path``

    Returns
    -------
        Updated value
    """
    if attrs.has(type(value)):
        return insert_path_prefix(cast(AttrsInstance, value), prefix)

    if isinstance(value, Path):
        return prefix / value

    return value


def get_step_config_ids(step_configs: Iterable[NotebookConfigLike]) -> list[str]:
    """
    Get available config IDs from an iterable of step configurations

    Parameters
    ----------
    step_configs
        Step configurations from which to retrieve the step config IDs

    Returns
    -------
        Step config ID from each step config in ``step_configs``

    Raises
    ------
    AttributeError
        Any object in ``step_configs`` is missing a `"step_config_id"` attribute.
    """
    attr_to_get = "step_config_id"
    try:
        res = [getattr(c, attr_to_get) for c in step_configs]
    except AttributeError as exc:
        # A bit of mucking around, but provides much more information in case
        # we hit this error.
        for c in step_configs:
            if not hasattr(c, attr_to_get):
                msg = f"{c!r} is missing a `{attr_to_get}` attribute"
                raise AttributeError(msg) from exc

        raise NotImplementedError("How did you get here") from exc  # pragma: no cover

    return res


def get_config_for_step_id(
    config: Any,
    step: str,
    step_config_id: str,
) -> Any:
    """
    Get configuration for a specific value of step config ID for a specific step

    This will fail if ``step`` isn't a part of ``config``

    Parameters
    ----------
    config
        Config from which to retrieve the step config

    step
        Step in ``config`` in which to find a step that has the a step config ID
        equal to ``step_config_id``

    step_config_id
        Value that the retrieved configuration's ``step_config_id`` must match

    Returns
    -------
        Configuration for step ``step`` in ``config`` with a step config ID
        equal to ``step_config_id``

    Raises
    ------
    ValueError
        No step could be found with ID equal to ``step_config_id``

    AttributeError
        ``config`` does not have a an attribute equal to the value given in
        ``step``
    """
    possibilities: Iterable[NotebookConfigLike] = getattr(config, step)
    available_step_config_ids = []
    for poss in possibilities:
        if poss.step_config_id == step_config_id:
            return poss
        else:
            available_step_config_ids.append(poss.step_config_id)

    raise ValueError(  # noqa: TRY003
        f"Couldn't find {step_config_id=} for {step=}. "
        f"Available step config IDs: {available_step_config_ids}"
    )


AI = TypeVar("AI", bound=AttrsInstance)
CB = TypeVar("CB", bound=ConfigBundleLike[Any])


def load_hydrate_write_config_bundle(
    configuration_file: Path,
    load_configuration_file: Callable[[Path], AI],
    create_config_bundle: ConfigBundleCreator[AI, CB],
    root_dir_output_run: Path,
    converter: Converter,
) -> CB:
    """
    Load, hydrate and write (to disk) :obj:`ConfigBundleLike`

    Parameters
    ----------
    configuration_file
        File from which to load the configuration

    load_configuration_file
        Callable to use to load the configuration from a file

    create_config_bundle
        Callable to use to create the :obj:`ConfigBundleLike` from the
        loaded configuration, the path in which the hydrated config will
        be written and the run's output root directory.

    root_dir_output_run
        Root directory in which to write output for this run

    converter
        Converter to use to serialise the output :obj:`ConfigBundleLike` to
        disk

    Returns
    -------
        Loaded :obj:`ConfigBundleLike`
    """
    config = load_configuration_file(configuration_file)
    config = insert_path_prefix(config, prefix=root_dir_output_run)
    config_bundle = create_config_bundle(
        config_hydrated=config,
        config_hydrated_path=root_dir_output_run / configuration_file.name,
        root_dir_output_run=root_dir_output_run,
    )
    write_config_in_config_bundle_to_disk(config_bundle=config_bundle, converter=converter)

    return config_bundle
